from KlebLib import fraction, polynomial, baseconversion, universaladdition, series
from typing import Any

class Test:
    def __init__(self, testType:str, **kwargs):
        self.testType = testType
        self.kwargs = kwargs

    def test(self, *args:Any) -> str:
        if self.testType == 'fraction':
            return self._fraction_test(args)
        elif self.testType == 'polynomial':
            return self._polynomial_test(args)
        elif self.testType == 'baseconversion':
            return self._conversion_test(args)
        elif self.testType == 'universaladdition':
            return self._universal_addition_test(args)
        elif self.testType == 'series':
            return self._series_test(args)

    def __repr__(self):
        output = f'Test({self.testType}, '

        #print(f'kwargs are of length {len(self.kwargs)}') #debug
        for i, (k, v) in enumerate(self.kwargs.items()):
            output += f'{k}=\'{v}\''
            if i != len(self.kwargs.keys()) - 1:
                output += ', '

        output += ')'
        return output

    #Individual tests

    def _fraction_test(self, args):
        fraction1 = fraction.Fraction(self.kwargs['fraction1'])
        fraction2 = fraction.Fraction(self.kwargs['fraction2'])

        opType = self.kwargs['opType']
        if opType == 'add' or opType == '+':
            fraction3 = fraction1 + fraction2
        elif opType == 'subtract' or opType == '-':
            fraction3 = fraction1 - fraction2
        elif opType == 'multiply' or opType == '*':
            fraction3 = fraction1 * fraction2
        elif opType == 'divide' or opType == '/':
            fraction3 = fraction1 / fraction2

        return str(fraction3)

    def _polynomial_test(self, args):
        testPolynomial = polynomial.Polynomial(self.kwargs['polynomial'])

        differentiated = testPolynomial.differentiate(args[0])
        integrated = testPolynomial.integrate(args[0])

        return str({
            'polynomial': testPolynomial.polynomial,
            'readable': str(testPolynomial),
            'differentiated': differentiated.polynomial,
            'integrated': integrated.polynomial
        })

    def _conversion_test(self, args):
        num = self.kwargs['num']
        base = self.kwargs['base']
        convertedBase = self.kwargs['baseToConvertTo']

        return baseconversion.convert_base(num, base, convertedBase)

    def _universal_addition_test(self, args):
        num1 = self.kwargs['num1']
        num2 = self.kwargs['num2']
        base = self.kwargs['base']

        return universaladdition.add(base, num1, num2)

    def _series_test(self, args):
        testSeries = series.Series(self.kwargs['data'], self.kwargs['type'], self.kwargs['strictType'])
        return str({'base series': str(testSeries),
                    'base series + arg': str(testSeries + args[0])})


import re
from typing import Union

class Polynomial:
    def __init__(self, polynomial:Union[list, str]):
        #print(f'the polynomials are {polynomial} and are of type {type(polynomial).__name__}') #debug
        if type(polynomial) is list:
            for term in polynomial:
                if not 'num' in term:
                    raise KeyError('Every term must contain key \'num\'')
            self.polynomial = polynomial
        else:
            self.polynomial = self._parse(polynomial, self.get_variables(polynomial))
        #print(self.polynomial) #debug

    def _parse(self, polynomial:str, variables:list) -> list:
        #parse the polynomial into a list of dictionaries
        
        #print(f'parsing polynomial {polynomial}') #debug
        output = []
        negatives = []
        additions = re.finditer(r'\s+[+-]\s+', polynomial)
        startNegative = re.search(r'^-', polynomial)
        if startNegative:
            negatives.append(0)
            polynomial = polynomial[1:]

        for i, match in enumerate(additions):
            #print(f'the match at position {i} is {match.group()}') #debug
            if match.group() == ' - ':
                negatives.append(i + 1)

        terms = re.split(r'\s+[+-]\s+', polynomial)
        #print(f'the terms before translation are {terms}') #debug
        terms = [self._translate_super(i) for i in terms]
        #print(f'the terms after translation are {terms}') #debug
        #print(f'the negatives are {negatives}') #debug

        for i, term in enumerate(terms):
            term = term.strip()
            #print(f'parsing term {term} at position {i}') #debug
            
            currentTerm = {}
                
            #Check if term should be negative
            if i in negatives:
                negativeMultiple = -1
            else:
                negativeMultiple = 1
                        
            for variable in self.get_variables(term):
                if f'{variable}^' in term:
                    currentTerm[variable] = self._trim_num(term, 'right')
                            
                else: #If it has no exponent
                    currentTerm[variable] = 1

            trimmed = self._trim_num(term, 'left')
            if trimmed: #If there is a coefficient
                currentTerm['num'] = trimmed * negativeMultiple
            else:
                currentTerm['num'] = negativeMultiple

            if currentTerm:
                output.append(currentTerm)

        return output

    def differentiate(self, varToDiff:str=None):
        if varToDiff is None:
            #print(f'attempting to imply variable from polnomial with variables {self.get_variables(self.polynomial)}') #debug
            if len(self.get_variables(self.polynomial)) != 1:
                raise TypeError('cannot implicitly detect variable for polynomials of multiple variables')
            else:
                varToDiff = self.get_variables(self.polynomial)[0]
        
        outputPolynomial = []
        for term in self.polynomial:
            termToEdit = {}
            edited = False
            for variable, exponent in term.items():
                if variable == varToDiff:
                    termToEdit['num'] = term['num'] * exponent
                    if exponent - 1 != 0:
                        termToEdit[variable] = exponent - 1
                    edited = True
                elif variable != 'num':
                    termToEdit[variable] = exponent
            if not edited:
                continue

            outputPolynomial.append(termToEdit)

        return Polynomial(outputPolynomial)

    def integrate(self, varToIntegrate:str=None):
        if varToIntegrate is None:
            if len(self.get_variables(self.polynomial)) != 1:
                raise TypeError('cannot implicitly detect variable for polynomials of multiple variables')
            else:
                varToIntegrate = self.get_variables(self.polynomial)[0]
                
        outputPolynomial = []
        for term in self.polynomial:
            termToEdit = {}
            edited = False
            for variable, exponent in term.items():
                if variable == varToIntegrate:
                    termToEdit['num'] = term['num'] / (exponent + 1)
                    termToEdit[variable] = exponent + 1
                    edited = True
                elif variable != 'num':
                    termToEdit[variable] = exponent
            if not edited:
                termToEdit['num'] = term['num']
                termToEdit[varToIntegrate] = 1

            outputPolynomial.append(termToEdit)

        return Polynomial(outputPolynomial)

    def integrate_definite(self, varToIntegrate:str, max:int, min:int):
        integrated = self.integrate(varToIntegrate)
        outputs = [[], []]
        limits = [max, min]

        for i, output in enumerate(outputs):
            for term in integrated.polynomial:
                outputTerm = {}
                edited = False
                for variable, exponent in term.items():
                    if variable == varToIntegrate:
                        outputTerm['num'] = term['num'] * limits[i] ** exponent
                        edited = True
                    elif variable != 'num':
                        outputTerm[variable] = exponent

                if not edited:
                    outputTerm['num'] = term['num']

                output.append(outputTerm)

            #print(f'output {i} before cleanup is {output}') #debug
            output = self._consolidate(output)
            #print(f'output {i} after cleanup is {output}') #debug

        upper = Polynomial(outputs[0])
        lower = Polynomial(outputs[1])

        return upper - lower

    def __add__(self, other):
        #print(f'adding polynomials {self} and {other}') #debug
        outputPolynomial = self.polynomial.deepcopy()
        
        for i, term in enumerate(other.polynomial.copy()):
            #print(f'adding term {term} to polynomial {outputPolynomial}') #debug
            location = self._locate(outputPolynomial.copy(), term.copy())
            if not type(location) is bool:
                outputPolynomial[location]['num'] += term['num']
            else:
                outputPolynomial.append(term.copy())

        #print(f'finished polynomial is {self.consolidate(outputPolynomial)}') #debug
        return Polynomial(self._consolidate(outputPolynomial))

    def __sub__(self, other):
        #print(f'subtracting polynomial {other} from {self}') #debug
        outputPolynomial = self.polynomial.deepcopy()
        
        for i, term in enumerate(other.polynomial.copy()):
            #print(f'subtracting term {term} from polynomial {outputPolynomial}') #debug
            location = self._locate(outputPolynomial.copy(), term.copy())
            if not type(location) is bool:
                outputPolynomial[location]['num'] -= term['num']
            else:
                currentTerm = {}
                for variable, exponent in term.items():
                    if variable == 'num':
                        currentTerm['num'] = -exponent
                    else:
                        currentTerm[variable] = exponent
                outputPolynomial.append(currentTerm)

        #print(f'finished polynomial is {self.consolidate(outputPolynomial)}') #debug
        return Polynomial(self.consolidate(outputPolynomial))

    def __iadd__(self, other):
        #print(f'adding polynomials {self} and {other}') #debug
        outputPolynomial = self.polynomial.deepcpy()
        
        for i, term in enumerate(other.polynomial.copy()):
            #print(f'adding term {term} to polynomial {outputPolynomial}') #debug
            location = self._locate(outputPolynomial.copy(), term.copy())
            if not type(location) is bool:
                outputPolynomial[location]['num'] += term['num']
            else:
                outputPolynomial.append(term.copy())

        #print(f'finished polynomial is {self.consolidate(outputPolynomial)}') #debug
        self.polynomial = self.consolidate(outputPolynomial)
        return self

    def __isub__(self, other):
        #print(f'subtracting polynomial {other} from {self}') #debug
        outputPolynomial = self.polynomial.deepcopy()
        
        for i, term in enumerate(other.polynomial.copy()):
            #print(f'suntracting term {term} from polynomial {outputPolynomial}') #debug
            location = self._locate(outputPolynomial.copy(), term.copy())
            if not type(location) is bool:
                outputPolynomial[location]['num'] -= term['num']
            else:
                currentTerm = {}
                for variable, exponent in term.items():
                    if variable == 'num':
                        currentTerm['num'] = -exponent
                    else:
                        currentTerm[variable] = exponent
                outputPolynomial.append(currentTerm)

        #print(f'finished polynomial is {self.consolidate(outputPolynomial)}') #debug
        self.polynomial = self.consolidate(outputPolynomial)
        return self

    def __str__(self):
        output = ''

        i = 0
        for term in self.polynomial:
            if term['num'] < 0:
                if term['num'] != -1:
                    if i == 0:
                        output += f'-{self._int_if_pos(abs(term["num"]))}'
                    else:
                        output += f' - {self._int_if_pos(abs(term["num"]))}'
                else:
                    if i == 0:
                        output += '-'
                    else:
                        output += ' + '
            elif i != 0:
                if term['num'] == 1:
                    output += ' + '
                else:
                    output += f' + {self._int_if_pos(term["num"])}'
            elif term['num'] != 1:
                output += str(self._int_if_pos(term['num']))

            i += 1
                
            for variable, exponent in term.items():
                if variable != 'num':
                    if exponent == 1:
                        output += str(variable)
                    else:
                        output += variable + self._super(self._int_if_pos(exponent))
                
        return output

    def __repr__(self):
        return f'polynomial.Polynomial({self.polynomial})'

    def __float__(self):
        temp = self.polynomial[0].copy()
        del temp['num']
        
        if len(self.polynomial) == 1 and not temp:
            return float(self.polynomial[0]['num'])
        else:
            raise ValueError('cannot convert polynomial with variables to float')

    def __int__(self):
        try:
            return int(float(self))
        except ValueError:
            raise ValueError('cannot convert polynomial with variables to int')

    def __getitem__(self, variable):
        output = []
        
        for term in self.polynomial:
            if variable in term:
                output.append(term.copy())

        return Polynomial(output)

    def __eq__(self, other):
        equal = True

        for term in self.polynomial:
            #print(f'comparing term {term} from self') #debug
            if term not in other.polynomial:
                #print(f'didn\'t find term {term} in other') #debug
                equal = False
                
        for term in other.polynomial:
            #print(f'comparing term {term} from other') #debug
            if term not in self.polynomial:
                #print(f'didn\'t find term {term} in self') #debug
                equal = False

        return equal

    def _int_if_pos(self, num:Union[float, str]) -> Union[int, float, str]:
        #print(f'checking if {num} can be int') #debug
        try:
            assert int(num) == num
        except AssertionError:
            return num
        except ValueError:
            return num
        else:
            return int(num)

    def get_variables(self, polynomial:Union[list, str]) -> list:
        variables = set()

        if type(polynomial) is str:
            for character in polynomial:
                #If it is a letter
                if (ord(character) >= 65 and ord(character) <= 90) or (ord(character) >= 97 and ord(character) <= 122):
                    variables.add(character)

        elif type(polynomial) is list:
            for term in polynomial:
                for variable in term:
                    if variable != 'num':
                        variables.add(variable)

        return list(variables)

    def _trim_num(self, string:str, side:str) -> float:
        #print(f'trimming {string}') #debug
        numPos = self._get_num_pos(string, side)

        if numPos:
            #print(f'num found between positions {numPos.start()} and {numPos.end()}') #debug
            return float(string[numPos[0]:numPos[1]])
        else:
            #print(f'num not found') #debug
            return 0.0

    def _get_num_pos(self, string:str, side:str) -> Union[list, None]:
        if side == 'left':
            numPos = re.search(r'^-?\d+\.?\d*', string)
        else:
            numPos = re.search(r'-?\d+\.?\d*$', string)

        if numPos:
            return [numPos.start(), numPos.end()]
        else:
            return None

    def _locate(self, polynomial:list, searchTerm:dict) -> Union[int, bool]:
        #print(f'locating term {searchTerm} in polynomial {polynomial}') #debug
        del searchTerm['num']
        for i, term in enumerate(polynomial):
            termToEdit = term.copy()
            del termToEdit['num']
            if searchTerm == termToEdit:
                #print(f'found term at position {i}') #debug
                return i
        #print('didn\'t find term') #debug
        return False

    def _consolidate(self, polynomial:list) -> list:
        output = polynomial.copy()
        termsToRemove = []
        
        total = 0
        for term in output:
            numVars = 0
            for variable in term:
                if variable != 'num':
                    numVars += 1

            if numVars == 0:
                total += term['num']
                termsToRemove.append(term)

        if total:
            output.append({'num': total})

        for term in termsToRemove:
            del output[output.index(term)]

        return output

    def _copy(self, listToCopy:list) -> list:
        output = []
        
        for item in listToCopy:
            output.append(item.copy())

        return output

    def _super(self, num:str) -> str:
        superscriptMap = {'1': '¹', '2': '²', '3': '³', '4': '⁴', '5': '⁵', '6': '⁶', '7': '⁷', '8': '⁸', '9': '⁹', '-': '⁻'}

        return self._translate(str(num), superscriptMap)

    def _translate_super(self, string:str) -> str:
        #print(f'translating {string}') #debug
        superscriptMap = {'1': '¹', '2': '²', '3': '³', '4': '⁴', '5': '⁵', '6': '⁶', '7': '⁷', '8': '⁸', '9': '⁹', '-': '⁻'}

        superscriptMap = {v:k for (k, v) in superscriptMap.items()}

        string = self._translate(string, superscriptMap)

        num = self._trim_num(string, 'right')
        numPos = self._get_num_pos(string, 'right')

        if numPos:
            if numPos[0] == 0:
                return string
            if string[numPos[0] - 1] == '^':
                return string[:numPos[0]] + str(num)
            else:
                return string[:numPos[0]] + '^' + str(num)
        else:
            return string
        
    def _translate(self, string:str, table:dict) -> str:
        trans = str.maketrans(
            ''.join(table.keys()),
            ''.join(table.values())
        )

        return string.translate(trans)


import re
from typing import Union

class Fraction:
    def __init__(self, fraction:Union[list, str, int, float]):
        if type(fraction) is list:
            self.num = fraction[0]
            self.dem = fraction[1]
            
        elif type(fraction) is str:
            fractionNums = fraction.split('/')
            fractionNums = [int(i) for i in fractionNums]
            self.num = fractionNums[0]
            self.dem = fractionNums[1]

        elif type(fraction) is int or type(fraction) is float:
            fractionNums = self._num_to_fraction(fraction)
            self.num = fractionNums[0]
            self.dem = fractionNums[1]
            
        else:
            raise TypeError(f'cannot parse type {type(fraction).__name__}')

        self._simplify()

    #Get the greatest common divisor of the numerator and denominator
    @property
    def GCD(self) -> int:
        num1 = self.num
        num2 = self.dem

        if num1 < num2:
            temp = num1
            num1 = num2
            num2 = temp

        remainder = num1 % num2
        while remainder != 0:
            num1 = num2
            num2 = remainder

            remainder = num1 % num2

        #print(f'GCD is {abs(num2)}') #debug
        return abs(num2)

    #Simplify the fraction
    def _simplify(self) -> None:
        #print(f'simplifying {self}') #debug
        num = self.num
        dem = self.dem

        #Ensure that negatives are represented in the numerator and there is no double negative
        if dem < 0:
            #print('flipping negatives') #debug
            num = -num
            dem = -dem

        num //= self.GCD
        dem //= self.GCD

        self.num = num
        self.dem = dem

    #Simplify a given fraction in list form
    def _simplify_nums(nums):
        return Fraction(nums, True).nums

    #Automatically simplify fractions
    def autosimplify(func):
        def inner(*args, **kwargs):
            fraction = func(*args, **kwargs)
            fraction.simplify
            return fraction
        return inner

    #Output the numbers as a fraction
    def __str__(self) -> str:
        return (f'{self.num}/{self.dem}')

    def __repr__(self) -> str:
        return f'fraction.Fraction([{self.num}, {self.dem}])'

    #Output the fraction as an int
    def __int__(self) -> int:
        return int(self.num / self.dem)

    #Output the fraction as a float
    def __float__(self) -> float:
        return self.num / self.dem

    #Compares this fraction to another
    def __eq__(self, other):
        return self.num == other.num and self.dem == other.dem

    #Compares this fraction to another and inverts
    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return float(self) > float(other)

    def __ge__(self, other):
        return float(self) >= float(other)

    def __lt__(self, other):
        return float(self) < float(other)

    def __le__(self, other):
        return float(self) <= float(other)

    #Adds two given fractions
    def _add(self, fraction1, fraction2):
        num = (fraction1.num * fraction2.dem) + (fraction1.dem * fraction2.num)
        dem = fraction1.dem * fraction2.dem

        return Fraction([num, dem])

    #Subtracts two given fractions
    def _sub(self, fraction1, fraction2):
        num = (fraction1.num * fraction2.dem) - (fraction1.dem * fraction2.num)
        dem = fraction1.dem * fraction2.dem

        return Fraction([num, dem])

    #Multiplies two given fractions
    def _mul(self, fraction1, fraction2):
        num = fraction1.num * fraction2.num
        dem = fraction1.dem * fraction2.dem

        return Fraction([num, dem])

    #Divides two given fractions
    def _truediv(self, fraction1, fraction2):
        num = fraction1.num * fraction2.dem
        dem = fraction1.dem * fraction2.num

        return Fraction([num, dem])

    #Adds another fraction to this one and returns a new fraction
    def __add__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
            
        return self._add(self, other)

    #Subtracts another fraction from this one and returns a new fraction
    def __sub__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        return self._sub(self, other)

    #Multiplies another fraction by this one and returns a new fraction
    def __mul__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        return self._mul(self, other)

    #Divides this fraction by another one and returns a new fraction
    def __truediv__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        return self._mul(self, other)

    #Adds another fraction to this one and returns a new fraction
    def __radd__(self, number):
        #print(f'adding self to {number}') #debug
        other = self._num_to_fraction(number)
            
        return self._add(other, self)

    #Subtracts another fraction from this one and returns a new fraction
    def __rsub__(self, number):
        #print(subtracting self from {number}') #debug
        other = self._num_to_fraction(number)
            
        return self._sub(other, self)

    #Multiplies another fraction by this one and returns a new fraction
    def __rmul__(self, number):
        #print(multiplying {number} by self') #debug
        other = self._num_to_fraction(number)
        
        return self._mul(other, self)

    #Divides this fraction by another one and returns a new fraction
    def __rtruediv__(self, number):
        #print(dividing {number} by self') #debug
        other = self._num_to_fraction(number)
        
        return self._truediv(other, self)

    #Adds this fraction to another one and replaces this fraction with the result
    def __iadd__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        answer = self._add(self, other)
        
        self.num = answer.num
        self.dem = answer.dem
        return self

    #Subtracts another fraction from this one and replaces this fraction with the result
    def __isub__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        answer = self._sub(self, other)
        
        self.num = answer.num
        self.dem = answer.dem
        return self

    #Multiplies this fraction by another one and replaces this fraction with the result
    def __imul__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        answer = self._mul(self, other)
        
        self.num = answer.num
        self.dem = answer.dem
        return self

    #Divides this fraction by another one and replaces this fraction with the result
    def __itruediv__(self, other):
        if type(other) is not Fraction:
            try:
                other = self._num_to_fraction(other)
            except TypeError:
                raise TypeError(f'Cannot add type {type(other)} to fraction')
                
        answer = self._truediv(self, other)
        
        self.num = answer.num
        self.dem = answer.dem
        return self

    #Flips this fraction between positive and negative
    def __neg__(self):
        return Fraction([-self.num, self.dem])

    #Updates this fraction to its reciprocal
    def invert(self):
        return Fraction([self.dem, self.num])

    #Returns either the numerator or denominator, as if the fraction were a list in form [num, dem]
    def __getitem__(self, index):
        if index == 0:
            return self.num
        elif index == 1:
            return self.dem
        else:
            raise IndexError(f'Invalid value for index: {index}. Must be 0 or 1')
    
    #Returns a mixed fraction
    @property
    def mixed(self):
        integerPart = 0
        num = self.num
        dem = self.dem

        while num > dem:
            num -= dem
            integerPart += 1

        return [integerPart, [num, dem]]

    #Sets the fraction to the top-heavy form of a given mixed fraction
    @mixed.setter
    def mixed(self, fraction):
        #Fraction given in form [integerPart, [num, dem]]
        fraction[1][0] += fraction[0] * fraction[1][1]
        self.num = fraction[1][0]
        self.dem = fraction[1][1]

    #Returns a list of the numerator and denominator
    @property
    def nums(self):
        return ([self.num, self.dem])

    #Updates the fraction through the use of nums
    @nums.setter
    def nums(self, numsList:list):
        self.num = numsList[0]
        self.dem = numsList[1]

        self.simplify()

    def _num_to_fraction(self, number):
        if round(number, 0) == number:
            return Fraction([number, 1])
        else:
            decPlaces = len(str(number)[re.search(r'\.', str(number)).start() + 1:])
            number *= 10 ** decPlaces

            return Fraction([number, 10 ** decPlaces])


def ceiling(num:float, decPlaces:int=0) -> float:
    #Check validity of input
    if not isinstance(num, int) and not isinstance(num, float):
        raise ValueError(f'Expected int or float, got {type(num).__name__}')
    if not isinstance(decPlaces, int):
        raise ValueError(f'Expected int, got {type(decPlaces).__name__}')
    
    #Set up variables
    overflow = False
    num = str(num)
    numDigits = [i for i in num]
    decPoint = numDigits.index('.')

    #Removes the decimal point and finds the last decimal place that will remain
    del numDigits[decPoint]
    lastPlace = decPoint + decPlaces - 1

    #Turns all of the digits into int
    for i, digit in enumerate(numDigits):
        numDigits[i] = int(digit)

    #Rounds up
    complete = False
    roundMod = 0
    while not complete:
        if numDigits[lastPlace - roundMod] < 9:
            #If 1 can be added to the selected position
            numDigits[lastPlace - roundMod] += 1
            complete = True
        else:
            if numDigits[lastPlace] - roundMod - 1 < 0:
                numDigits.insert(0, 1)
                complete = True
                overflow = True
            else:
                numDigits[lastPlace - roundMod] = 0
                roundMod += 1

    #Remove the extra digits
    while len(numDigits) > lastPlace + 1:
        del numDigits[-1]

    #Reformats and outputs the answer
    numDigits.insert(decPoint + int(overflow), '.')
    num = ''
    for i in numDigits:
        num += str(i)
        
    return float(num)
    
def smartRound(num:float, decPlaces:int=0) -> float:
    if str(num)[decPlaces + 1] >= 5:
        return ceiling(num, decPlaces)
    else:
        return round(num, decPlaces)


from typing import Any

class Series:
    def __init__(self, item:Any, selfType:type=None, subType:type=None):
        #print(f'creating series from item {item}') #debug
        skip = False

        if type(item) is selfType:
            #print('using given type') #debug
            self.value = item
            self.type = selfType
            self.next = None
            skip = True
            
        elif type(item) is set or type(item) is tuple:
            #print(f'converting item from {type(item)} to list') #debug
            item = list(item)
            
        elif type(item) is not list:
            if selfType and type(item) is not selfType:
                raise TypeError(f'type of given item {item} is not the same as given type {selfType}')
                
            self.value = item
            self.type = type(item)
            self.next = None
            
        if (type(item) is list) and (not skip):
            if selfType and type(item[0]) is not selfType:
                raise TypeError(f'type of given item {item[0]} is not the same as given type {selfType}')
                
            for subItem in item:
                if type(subItem) is not type(item[0]):
                    raise TypeError('container passed to series must be of uniform type')
                    
            self.type = type(item[0])
            self.value = item[0]
            if len(item) > 1:
                self.next = Series(item[1:])
            else:
                self.next = None

        if subType is not None:
            if selfType is not Series:
                raise TypeError('series type must be series if subtype is given')

            self.subType = subType
            for series in self:
                if series.type != subType:
                    raise TypeError(f'series type {series.type} must be equal to subtype {subType}')
        else:
            self.subType = None

    def __len__(self):
        if self.value is not None:
            length = 1
            current = self
            while current.next is not None:
                length += 1
                current = current.next

            return length
            
        else:
            return 0

    @property
    def objects(self):
        #print(f'getting objects of series with head {self.value}') #debug
        result = [self]
        current = self
        while current.next is not None:
            #print(f'value of current object is {current.value}') #debug
            current = current.next
            result.append(current)

        #print('got objects') #debug
        return result

    def __getitem__(self, index:int):
        if index >= len(self) or -index > len(self):
            raise IndexError('series index out of range')

        return self.objects[index].value

    def __setitem__(self, index:int, value):
        if index >= len(self) or -index > len(self):
            raise IndexError('series index out of range')

        self.objects[index].value = value

    def __delitem__(self, index):
        if index >= len(self) or -index > len(self):
            raise IndexError('series index out of range')
            
        self.objects[index-1].next = self.objects[index+1]

    def insert(self, index, value):
        temp = self.objects[index]
        self.objects[index-1].next = Series(value)
        self.objects[index].next = temp
        del temp

    def __iter__(self):
        self.iterIndex = -1
        return self

    def __next__(self):
        self.iterIndex += 1
        if self.iterIndex >= len(self):
            raise StopIteration
            
        return self[self.iterIndex]

    def __add__(self, other):
        result = self.deepcopy()

        if type(other) is self.type:
            if result.subType is not None and type(other) is not result.subType:
                raise TypeError(f'series of type {other.type} cannot be appended to series of subtype {self.subType}')
                
            result.objects[-1].next = Series(other, type(other))
            return result

        elif type(other) is list or type(other) is tuple or type(other) is set:
            other = Series(other)
        
        if type(other) is Series:
            if self.type != other.type:
                raise TypeError(f'cannot add series of type {other.type} to series of type {self.type}')

            result.objects[-1].next = other
            
        else:
            raise TypeError(f'cannot add object of type {type(other)} to series of type {self.type}')

        return result

    def __radd__(self, other):
        if type(other) is self.type:
            output = Series(other)
            output.next = self.deepcopy()
            
        elif type(other) is list or type(other) is tuple or type(other) is set:
            output = Series(other)
            if output.type != self.type:
                raise TypeError(f'cannot add series of type {self.type} to container containing type {other.type}')

            output.objects[-1].next = self.deepcopy()
            
        else:
            raise IndexError(f'cannot add series of type {self.type} to object of type {type(other)}')

        return output

    def __iadd__(self, other):
        if type(other) is self.type:
            self.objects[-1].next = Series(other, type(other))
            return self
        
        elif type(other) is list or type(other) is tuple or type(other) is set:
            other = Series(other)
            
        if type(other) is Series:
            if self.type != other.type:
                raise TypeError(f'cannot add series of type {other.type} to series of type {self.type}')

            self.objects[-1].next = other
            
        else:
            raise TypeError(f'cannot add object of type {type(other)} to series of type {self.type}')

        return self

    def __str__(self):
        #print(f'coverting series with head value {self.value} to str') #debug
        output = '<'
        for i, item in enumerate(self):
            output += str(item)
            
            if i != len(self) - 1:
                output += ', '

        output += '>'
        return output

    def __repr__(self):
        output = 'series.Series(['
        for i, item in enumerate(self):
            output += repr(item)
            
            if i != len(self) - 1:
                output += ', '

        if self.type is Series:
            output += f'], series.Series'
        else:
            output += f'], {self.type.__name__}'
            
        if self.subType is not None:
            if self.subType is Series:
                output += ', series.Series'
            else:
                output += f', {self.subType.__name__}'

        output += ')'

        return output

    def __list__(self):
        return [item for item in self]

    def __set__(self):
        return set(list(self))

    def copy(self):
        output = Series(self[0], self.type, self.subType)
        current = output
        for i in range(1, len(self)):
            current.next = Series(self[i], self.type, self.subType)
            current = current.next

        return output

    def deepcopy(self):
        try:
            output = Series(self[0].deepcopy(), self.type, self.subType)
            
        except AttributeError:
            output = Series(self[0], self.type, self.subType)
            current = output
            for i in range(1, len(self)):
                current.next = Series(self[i])
                current = current.next
                
        else: #continue code in try block
            current = output
            for i in range(1, len(self)):
                current.next = Series(self[i].deepcopy(), self.type, self.subType)
                current = current.next

        return output


from math import log

_POSS_DIGITS = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz@#£&%;:€$¥_^§|~<>()[]{}.,!?ςερτυθιοσδφγξλζψωβцгшщзфлджэячиьбюъ'

def _convert_denary(num:str, base:int) -> str:
    #Set up variables
    decNum = 0
    decNumArr = []
    usedDigits = []
  
    #Create an array with the digits of the highest base
    if base > 10:
    	maxBase = base
    else:
    	maxBase = 10
  
    for i in range(maxBase):
    	usedDigits.append(_POSS_DIGITS[i])
   
    for digit in num:
    	if digit not in usedDigits:
      		raise ValueError(f'digit {digit} is not in base {base}')
   
    num = num[::-1]
    
    for i, digit in enumerate(num):
    	decNumArr.append(
      		usedDigits.index(digit) * (base ** i)
    	)
    
    decNumArr.reverse()
    
    for i in decNumArr:
    	decNum += i
    
    return decNum
  
def convert_base(num:str, base:int, ansBase:int) -> str:
    #Set up variables
    decNum = _convert_denary(num, base)
    usedDigits = []
    ansDigits = []
    ans = ''
  
    #Create an array with the digits of the highest base
    if base > ansBase:
    	maxBase = base
    else:
    	maxBase = ansBase
  
    for i in range(maxBase):
    	usedDigits.append(_POSS_DIGITS[i])
    
    for digit in num:
    	if digit not in usedDigits:
      		raise ValueError(f'digit {digit} is not in base {base}')

	#Fix the 'missing zero' error
    if not decNum:
        ansDigits = [0]
    
    #Succesive division of the denary number by the base with the remainder appended to the answer
    while decNum:
        ansDigits.append(decNum % ansBase)
        decNum //= ansBase
    
    ansDigits.reverse()
  
  #Converting each digit from denary to its value in the given base
    for i, digit in enumerate(ansDigits):
    	ansDigits[i] = usedDigits[digit]
    
    #Combining the digits into one string
    for digit in ansDigits:
    	ans += digit
  
    return ans

def _convert_dual_base_denary(num:str, outerBase:int, innerBase:int) -> str:
    #Convert a number from a dual base to denary
    innerSize = int(log(outerBase, innerBase))
    
    num = list(num)

    while len(num) % innerSize:
        num.insert(0, '0')
    
    #Splits the list into segments of length innerSize
    numArr = [num[i:i+innerSize] for i in range(0, len(num), innerSize)]
    
    decNumArr = []
    decNum = 0

    #Get the decimal values of all of the digits
    for innerBaseNum in numArr:
        decNumArr.append(_convert_denary(innerBaseNum, innerBase))
    decNumArr.reverse()

    #Add the decimal values to the answer, multiplying by the base raised to the power equal to the column number
    for i, num in enumerate(decNumArr):
        decNum += num * (outerBase ** i)

    return decNum

def convert_from_dual_base(num:str, outerBase:int, innerBase:int, ansBase:int) -> str:
    #Convert a number from a dual base to any other base
    decNum = _convert_dual_base_denary(num, outerBase, innerBase)
    ans = convert_base(str(decNum), 10, ansBase)
    return ans

def convert_to_dual_base(num:str, base:int, ansOuterBase:int, ansInnerBase:int) -> str:
    #Convert a number from any base to a dual base
    innerSize = int(log(ansOuterBase, ansInnerBase))
    ans = ''

    #Convert the number to the outer base
    outerBaseNum = convert_base(num, base, ansOuterBase)

    #Convert each digit to the inner base
    for digit in outerBaseNum:
        output = convert_base(digit, ansOuterBase, ansInnerBase)
        while len(output) < innerSize:
            output = '0' + output
        ans += output

    return ans

def convert_between_dual_bases(num:str, outerBase:str, innerBase:str, ansOuterBase:str, ansInnerBase:str) -> str:
    temp = convert_from_dual_base(num, outerBase, innerBase, 64)
    return convert_to_dual_base(temp, 64, ansOuterBase, ansInnerBase)


from KlebLib.baseconversion import *

def add(base:int, *nums:str) -> str:
    #Check validity of input
    if type(base) is not int:
        if type(base) is float:
            base = int(base)
        else:
            raise ValueError(f'Expected int, got {type(base).__name__}')
    
    #Set up nums
    nums = list(nums)
    if len(nums) < 2:
        raise ValueError('tuple nums must be of length 2 or greater')
  
    #print(f'{nums} start') #debug

	#Add the first 2 numbers, store the value as the first num and delete the second num
    while len(nums) > 2:
        #print('recursion activated') #debug
        nums[0] = add(base, nums[0], nums[1])
        del nums[1]
  
    #print(f'{nums} passed while loop') #debug
  
    num1 = nums[0]
    num2 = nums[1]
  
    #Set up variables
    carry = 0
    result = 0
    answer = []
    convertedAns = []
    finalAns = ''
    num1arr = []
    num2arr = []
  
    #Turn the strings into lists
    for (num1digit, num2digit) in zip(num1, num2):
        num1arr.append(num1digit)
        num2arr.append(num2digit)
    num1 = num1arr
    num2 = num2arr
  
    #Reverse the arrays and equalise the lengths
    num1.reverse()
    num2.reverse()
  
    while len(num1) != len(num2):
        if len(num1) > len(num2):
            num2.append('0')
    else:
      num1.append('0')
    
    #Convert nums to denary and add the numbers    
    for i, (num1digit, num2digit) in enumerate(zip(num1, num2)):
        num1digit = int(convert_base(num1digit, base, 10))
        num1[i] = num1digit
        num2digit = int(convert_base(num2digit, base, 10))
        num2[i] = num2digit
		
        result = num1digit + num2digit + carry
        if result >= carry:
        	carry = result // base
        	result = int(result % base)
        answer.append(result)
        #print(f'result: {result}') #debug
        #print(f'carry: {carry}') #debug
    
    #Handle overflow
    if carry > 0:
    	answer.append(carry)
    
    #Reverse answer to be in the correct order
    #print(f'reverse list answer: {answer}') #debug
    answer.reverse()
    #print(f'list answer: {answer}') #debug
  
    #Convert digits to string
    for i in range(len(answer)):
        answer[i] = str(answer[i])
    #print(f'string list answer: {answer}') #debug
 
    #Convert answer back to original base
    for i in range(len(answer)):
        convertedAns.append(convert_base(answer[i], 10, base))
    #print(f'converted answer: {convertedAns}') #debug
	
    #Combine digits into one string
    for i in convertedAns:
    	finalAns += str(i)
    
    return finalAns


class Tree:
    def __new__(cls, item, l=None, r=None, *args, **kwargs):
        if type(item) is Tree:
            return item
        else:
            return super(Tree, cls).__new__(cls, *args, **kwargs)
        
    def __init__(self, item, l=None, r=None):
        #print(f'creating tree from item {item}, left {l}, and right {r}') #debug
        
        if type(item) is dict and l is None and r is None:
            self.value = list(item.keys())[0]
            #print(f'self value is {self.value}') #debug
            sub = list(item.values())[0]
            
            if sub is None:
                self.l = None
                self.r = None
                
            else:
                keys = list(sub.keys())
                values = list(sub.values())
                #print(f'sub is {sub}, keys are {keys}, values are {values}') #debug
                
                self.l = Tree({
                    str(keys[0]): values[0]
                })
                #print(f'self.l is {self.l}') #debug

                self.r = Tree({
                    str(keys[1]): values[1]
                })
                #print(f'self.r is {self.r}') #debug
            
        else:
            self.value = item
            self.l = Tree(l)
            self.r = Tree(r)

    def __repr__(self):
        return self.value

    @property
    def max_depth(self):
        if self.l is not None:
            lDepth = self.l.max_depth
        else:
            lDepth = -1

        if self.r is not None:
            rDepth = self.r.max_depth
        else:
            rDepth = -1

        if lDepth > rDepth:
            return lDepth + 1
        else:
            return rDepth + 1

    @property
    def min_depth(self):
        if self.l is not None:
            lDepth = self.l.min_depth
        else:
            lDepth = -1

        if self.r is not None:
            rDepth = self.r.min_depth
        else:
            rDepth = -1

        if lDepth < rDepth:
            return lDepth + 1
        else:
            return rDepth + 1

    def __getitem__(self, index):
        if index == 0:
            return self.value
            
        elif type(index) is str:
            current = self
            for direction in index:
                #print(f'current is {self}') #debug
                if direction == '<':
                    current = current.l
                elif direction == '>':
                    current = current.r
    
                if current is None:
                    raise IndexError('tree index out of depth')
    
            return current.value

        else:
            raise KeyError(str(index))


from setuptools import setup

setup(
    name='KlebLib',
    version='1.0.0',
    author='Caleb Robson',
    author_email='robson.caleb.299@gmail.com',
    packages=['KlebLib'],
    url='http://pypi.python.org/pypi/KlebLib/',
    license='MIT',
    description='A collection of classes and functions that no-one will ever need.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown'
)
